#include <stdio.h>
#include <stdint.h>
#include <stdarg.h>
#include <limits.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <vector>
#include <mpi.h>

#include "solver.h"

template <class T> static inline void swap(T& x, T& y) { T t=x; x=y; y=t; }
template <class T> static inline void swap(T* a, int x, int y) { T t=a[x]; a[x]=a[y]; a[y]=t; }
#ifndef TIMER
#define TIMER
#define NS_PER_SEC 1000000000
double wall_time_diff(int64_t ed, int64_t st)
{
	return (double)(ed-st)/(double)NS_PER_SEC;
}
int64_t wall_clock_ns()
{
#if __unix__
	struct timespec tspec;
	clock_gettime(CLOCK_MONOTONIC, &tspec);
	return tspec.tv_sec*NS_PER_SEC + tspec.tv_nsec;
#else
#if __MACH__
	return 0;
#else
	struct timeval tv;
	gettimeofday( &tv, NULL );
	return tv.tv_sec*NS_PER_SEC + tv.tv_usec*1000;
#endif
#endif
}
#endif
double static inline ddot3(const double *x, const double *y, const double *z, int l)
{
#if 0
	double s = 0;
	for(int i=0; i<l; i++)
		s += (x[i] * y[i] * z[i]);
	return s;
#else
	double s = 0;
	int m = l-4;
	int i;
	for (i = 0; i < m; i += 5)
		s +=    x[i+0] * y[i+0] * z[i+0] +
			x[i+1] * y[i+1] * z[i+1] +
			x[i+2] * y[i+2] * z[i+2] +
			x[i+3] * y[i+3] * z[i+3] +
			x[i+4] * y[i+4] * z[i+4] ;

	for ( ; i < l; i++)        /* clean-up loop */
		s += x[i] * y[i] * z[i];

	return s;
#endif
}

#ifndef min
template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
#endif

#ifndef max
template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
#endif

#define MEPS 1e-16L
#define PIVOTING
#define DEBUG_LDL

// solve Ax=b , in which A[i*n+j] = Aij and we only solve submatrix A[m:m]
double solve_system(const double *_A, const double *_b, int n, int m, double *x)
{
	long double *A = new long double[n * n];
	for(int i=0; i<n*n; i++)
		A[i] = _A[i];
	long double *b = new long double[n];
	for(int i=0; i<n; i++)
		b[i] = _b[i];

#ifdef PIVOTING
	int *p = new int[m];
	for(int i=0; i<m; i++)
		p[i] = i;
#endif

	// decomp
	// A -> D0
	//      L00 D1
	//      L10 L11 D2
	for(int j=0; j<m; j++){
#ifdef PIVOTING
		int pj = j;
		long double pj_val = fabsl(A[j*n+j]);
		// find piv
		for(int i=j+1; i<m; i++){
			long double pi_val = fabsl(A[i*n+i]);
			if(pj_val < pi_val){
				pj_val = pi_val;
				pj = i;
			}
		}
		// swap rows and columns of j and pj
		for(int i=0; i<m; i++)
			swap(A, j*n+i, pj*n+i);
		for(int i=0; i<m; i++)
			swap(A, i*n+j, i*n+pj);
		swap(b, j, pj);
		swap(p, j, pj);
#endif

		// loop invariant: Aik = Lik, Akk = Dk with k<j
		long double Dj = A[j*n+j];
		for(int k=0; k<j; k++){
			Dj -= A[j*n+k]*A[j*n+k] * A[k*n+k];
		}
		A[j*n+j] = Dj;
		//assert(fabs(Dj) >= MEPS);
		if(fabsl(Dj) < MEPS){
			for(int i=j+1; i<m; i++)
				A[i*n+j] = 0;
			continue; // xj should be 0
		}

		for(int i=j+1; i<m; i++){
			long double Lij = A[i*n+j];
			for(int k=0; k<j; k++){
				Lij -= A[i*n+k] * A[j*n+k] * A[k*n+k];
			}
			Lij /= Dj;
			A[i*n+j] = Lij;
		}
	}
#if 0
	printf("L\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<=i; j++)
			printf("%+.3Le ", A[i*n+j]);
		printf("\n");
	}
#endif
	// invariant: A[i*n+j] = Lij, forall i>j
	//            A[i*n+i] = Dii, forall i

	// LD y = b
	// di yi =  bi - \sum_{j<i} yj Lij
	// place y in b
	for(int i=0; i<m; i++){
		// loop invariant: bk = yk forall k<i
		long double yi = b[i];
		for(int k=0; k<i; k++)
			yi -= A[i*n+k] * b[k];
		b[i] = yi;
	}
	for(int i=0; i<m; i++){
		long double Di = A[i*n+i];
		if(fabsl(Di) < MEPS)
			b[i] = 0;
		else
			b[i] /= Di;
	}
#if 0
	printf("y\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", b[i]);
	}
#endif
	// L'x = y
	// xi = yi - \sum{j>i} xj Lij
	for(int i=m-1; i>=0; i--){
		// loop invariant: xk ready forall k>i
		long double xi = b[i];
#ifdef PIVOTING
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[p[k]];
		x[p[i]] = (double)xi;
#else
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[k];
		x[i] = (double)xi;
#endif
	}
#if 0
	printf("\n");
	for(int i=0; i<n; i++)
		printf("d %.3e\n", x[i]);

	printf("A\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<n; j++)
			printf("%+.3e ", Acopy[i*n+j]);
		printf("\n");
	}
	printf("b\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", bcopy[i]);
	}
#endif

#if 1
	double err = 0;
	for(int i=0; i<m; i++){
		double Ax_i = 0;
		for(int j=0; j<m; j++)
			Ax_i += _A[i*n+j] * x[j];
		err += fabs(Ax_i-_b[i]);
	}
	double max_pv = 0;
	for(int i=0; i<m; i++)
		if(max_pv < _A[i*n+i])
			max_pv = (double)_A[i*n+i];
	if(max_pv != 0)
		err /= max_pv;
	//printf("err=%e\n", err);
	//assert(err < 1e-5);
#endif
	delete[] A;
	delete[] b;
#ifdef PIVOTING
	delete[] p;
#endif
	return err;
}

static int mpi_get_rank()
{
	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	return rank;	
}

static int mpi_get_size()
{
	int size;
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	return size;	
}

template<typename T>
void static mpi_allreduce(T *buf, const int count, MPI_Datatype type, MPI_Op op)
{
	size_t count_t = size_t(count);
	std::vector<T> buf_reduced(count_t);
	MPI_Allreduce(buf, buf_reduced.data(), count, type, op, MPI_COMM_WORLD);
	memcpy(buf, buf_reduced.data(), sizeof(T) * count_t);
}

#ifdef __cplusplus
extern "C" {
#endif

extern double dnrm2_(int *, double *, int *);
extern double ddot_(int *, double *, int *, double *, int *);
extern int daxpy_(int *, double *, double *, int *, double *, int *);
extern int dscal_(int *, double *, double *, int *);

#ifdef __cplusplus
}
#endif

static void default_print(const char *buf)
{
	fputs(buf,stdout);
	fflush(stdout);
}

void LBFGS::info(const char *fmt,...)
{
	if(mpi_get_rank()!=0)
		return;
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*lbfgs_print_string)(buf);
}

void LBFGS::set_print_string(void (*print_string) (const char *buf))
{
	lbfgs_print_string = print_string;
}

LBFGS::LBFGS(const function *fun_obj, double eps, int m, double eta, int max_iter)
{
	this->fun_obj=const_cast<function *>(fun_obj);
	this->eps=eps;
	this->max_iter=max_iter;
	this->eta = eta;
	lbfgs_print_string = default_print;
	this->M = m;
	int nr_node = mpi_get_size();
	int rank = mpi_get_rank();
	int n = this->fun_obj->get_nr_variable();
	int shift = int(ceil(double(n) / double(nr_node)));
	start = shift * rank;
	length = min(max(n - start, 0), shift);
	if (length == 0)
		start = 0;
	offset = 0;
}

LBFGS::~LBFGS()
{
}

void LBFGS::lbfgs(double *w)
{
	int nr_node = mpi_get_size();
	int n = fun_obj->get_nr_variable();
	int shift = int(ceil(double(n) / double(nr_node)));
	int i;
	int k = 0;
	double f;
	int search = 1, iter = 1, inc = 1;
	double minusone = -1;
	double *g;
	double *old_g;
	double init_size = 1;
	double *step = new double[n];
	double *local_step = new double[length];
	memset(w, 0, sizeof(double) * size_t(n));
	int64_t time_st = wall_clock_ns();
	f = fun_obj->fun(w);
	g = new double[n];
	fun_obj->grad(w, g);
	double gnorm1 = dnrm2_(&n, g, &inc);
	double gnorm = gnorm1;

	if (gnorm <= eps*gnorm1)
		search = 0;

	iter = 0;

	double **s = new double*[M];
	double **y = new double*[M];
	double **inner_product_matrix = new double*[2*M+1];
	double *rho = new double[M];
	double* delta = new double[M * 2 + 1];
	int *recv_count = new int[nr_node];
	int *displace = new int[nr_node];
	int counter = 0;
	for (i=0;i<nr_node;i++)
	{
		recv_count[i] = shift;
		displace[i] = counter;
		counter += shift;
		if (counter >= n)
		{
			counter = 0;
			shift = 0;
		}
		else if (counter + shift > n)
			shift = n - counter;
	}
	double *recv_buffer = new double[n];
	for (int i = 0; i < M; i++)
	{
		y[i] = new double[length];
		s[i] = new double[length];
	}
	for (int i=0; i < 2 * M + 1; i++)
		inner_product_matrix[i] = new double[2*M+1];

	double origDirDeriv = 0.0;
	double one = 1.0;
	double step_size = 0.0;
	double dTd = 0;
	int previous_M = 0;
	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed, time_st);
	info("iter %d m %d |g| %5.3e f %15.20e step %5.3e time %5.3e\n", iter, 0, gnorm, f, 0.0, tt);
	while (iter < max_iter && search)
	{
		int DynamicM = min(iter, M);
		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		time_st = wall_clock_ns();

		if (iter > 0)
		{
			g = new double[n];
			fun_obj->grad(w, g);
			double s0y0 = 0;

			memcpy(y[k], g + start, sizeof(double) * size_t(length));
			daxpy_(&length, &minusone, old_g + start, &inc, y[k], &inc);

			delete[] old_g;

			update_inner_products(inner_product_matrix, k, n, DynamicM, previous_M, s, y, g, delta, step_size);
			gnorm = sqrt(inner_product_matrix[2 * DynamicM][2 * DynamicM]);
			s0y0 = inner_product_matrix[2 * k][2* k + 1];
			info("iter %d m %d |g| %5.3e f %15.20e step %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, tt);
			if (gnorm <= eps*gnorm1)
				break;
			if (s0y0 == 0)
				break;
			rho[k] = 1.0 / s0y0;
			k = (k+1)%M;

			TwoLoop(inner_product_matrix, rho, DynamicM, k, n, delta);
			local_step = s[k];
			if (iter >= M)
			{
				dscal_(&length, delta + (2 * k), local_step, &inc);
				daxpy_(&length, delta + (2 * k + 1), y[k], &inc, local_step, &inc);
			}
			else
				memset(local_step, 0, sizeof(double) * size_t(length));
			daxpy_(&length, delta + (2 * DynamicM), g + start, &inc, local_step, &inc);
			for (i=0;i<k;i++)
			{
				daxpy_(&length, delta + (2 * i), s[i], &inc, local_step, &inc);
				daxpy_(&length, delta + (2 * i + 1), y[i], &inc, local_step, &inc);
			}
			for (i=k+1;i<DynamicM;i++)
			{
				daxpy_(&length, delta + (2 * i), s[i], &inc, local_step, &inc);
				daxpy_(&length, delta + (2 * i + 1), y[i], &inc, local_step, &inc);
			}
			MPI_Allgatherv(local_step, length, MPI_DOUBLE, step, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);

			origDirDeriv = 0;
			dTd = 0;
			int tmp_length = 2 * DynamicM + 1;
			origDirDeriv = ddot_(&tmp_length, delta, &inc, inner_product_matrix[2 * DynamicM], &inc);
			for (i=0;i<2 * DynamicM + 1;i++)
				dTd += delta[i] * ddot_(&tmp_length, delta, &inc, inner_product_matrix[i], &inc);
			init_size = 1;
		}
		else
		{
			local_step = s[0];
			memcpy(local_step, g + start, sizeof(double) * size_t(length));
			dscal_(&length, &minusone, local_step, &inc);
			memcpy(step, g, sizeof(double) * size_t(n));
			dscal_(&n, &minusone, step, &inc);
			init_size = 1.0 / gnorm;
			dTd = gnorm * gnorm;
			origDirDeriv = -gnorm * gnorm;
		}

		if (origDirDeriv >= 0)
		{
			info("WARNING: not a descent direction\n");
			break;
		}
		double ws = ddot_(&length, w + start, &inc, local_step, &inc);
		mpi_allreduce(&ws, 1, MPI_DOUBLE, MPI_SUM);

		step_size = fun_obj->line_search(step, origDirDeriv, dTd, ws, &f, eta, init_size);
		if (iter == 0)
		{
			inner_product_matrix[0][0] = dTd;
			delta[0] = -1;
		}
		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}
		dscal_(&length, &step_size, local_step, &inc);
		daxpy_(&length, &one, local_step, &inc, w + start, &inc);
		old_g = g;
		iter++;
		previous_M = DynamicM;
	}

	MPI_Allgatherv(w + start, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
	memcpy(w, recv_buffer, sizeof(double) * size_t(n));

	delete[] g;
	delete[] recv_count;
	delete[] recv_buffer;
	delete[] displace;
	delete[] step;
	for (int i = 0; i < M; i++)
	{
		delete[] y[i];
		delete[] s[i];
	}
	delete[] y;
	delete[] s;
	for (int i=0; i < 2 * M + 1; i++)
		delete[] inner_product_matrix[i];
	delete[] inner_product_matrix;
	delete[] rho;
	delete[] delta;

}

void LBFGS::update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, int previous_M, double** s, double** y, double* g, double* delta, double step_size)
{
	int i;
	int inc = 1;
	int prev_length = 2 * previous_M + 1;
	double *tmp = new double[prev_length + 1];
	double *tmp2 = new double[2 * DynamicM];

	double* buffer = new double[DynamicM * 2 + 2];
	memset(buffer, 0, sizeof(double) * size_t(DynamicM * 2 + 2));
	for (i=0;i<prev_length;i++)
		tmp[i] = ddot_(&prev_length, delta, &inc, inner_product_matrix[i], &inc);
	tmp[prev_length] = ddot_(&prev_length, delta, &inc, tmp, &inc) * step_size * step_size;
	for (i=0;i<prev_length;i++)
		tmp[i] *= step_size;

	for (i=0;i<DynamicM;i++)
	{
		buffer[2 * i] = ddot_(&length, g + start, &inc, s[i] + offset, &inc);
		buffer[2 * i + 1] = ddot_(&length, g + start, &inc, y[i] + offset, &inc);
	}
	buffer[2 * DynamicM] = ddot_(&length, g + start, &inc, g + start, &inc);
	buffer[2 * DynamicM + 1] = ddot_(&length, y[k] + offset, &inc, y[k] + offset, &inc);
	mpi_allreduce(buffer, 2 * DynamicM + 2, MPI_DOUBLE, MPI_SUM);
	for (i=0;i<prev_length - 1; i++)
		tmp2[i] = buffer[i] - inner_product_matrix[prev_length-1][i];

	tmp2[2 * k] = buffer[2 * k] - tmp[prev_length - 1];
//	s[k] * y[k] = s[k] * g - s[k] * old_g

	tmp2[2 * k + 1] = buffer[2 * DynamicM + 1];
//	y[k] * y[k]

	for (i=0;i< prev_length - 1;i++)
	{
		inner_product_matrix[2 * k][i] = tmp[i];
		inner_product_matrix[i][2 * k] = tmp[i];
	}

	for (i=0;i< 2 * DynamicM;i++)
	{
		inner_product_matrix[2 * k + 1][i] = tmp2[i];
		inner_product_matrix[i][2 * k + 1] = tmp2[i];
	}

	inner_product_matrix[2 * k][2 * k] = tmp[prev_length];

	for (i=0; i< 2 * DynamicM + 1; i++)
	{
		inner_product_matrix[2 * DynamicM][i] = buffer[i];
		inner_product_matrix[i][2 * DynamicM] = buffer[i];
	}

	delete[] buffer;
	delete[] tmp;
	delete[] tmp2;
}

void LBFGS::TwoLoop(double** inner_product_matrix, double* rho, int DynamicM, int k, int n, double *delta)
{
	int i, j;
	int inc = 1;
	int start = k-1;
	double beta;
	double *alpha = new double[DynamicM];
	int tmp_length = 2 * DynamicM + 1;

	memset(delta, 0, sizeof(double) * size_t(2 * DynamicM));
	memset(alpha, 0, sizeof(double) * size_t(DynamicM));
	delta[2 * DynamicM] = -1;
	if (k < DynamicM)
		start += DynamicM;
	int lastrho = start % DynamicM;

	for (i = 0; i < DynamicM; i++)
	{
		j = start % DynamicM;

		start--;
		if (rho[j] > 0)
		{
			alpha[j] = rho[j] * ddot_(&tmp_length, delta, &inc, inner_product_matrix[2 * j], &inc);
			delta[2 * j + 1] -= alpha[j];
		}
		else
		{
			fprintf(stderr,"ERROR: rho[%d] <= 0\n",i);
			return;
		}
	}
	if (lastrho != -1)
	{
		double scal = (1.0 / (rho[lastrho] * inner_product_matrix[2 * lastrho + 1][2 * lastrho + 1]));
		dscal_(&tmp_length, &scal, delta, &inc);
		for (i = 0; i < DynamicM; i++)
		{
			start++;
			j = start % DynamicM;
			beta = alpha[j] - rho[j] * ddot_(&tmp_length, delta, &inc, inner_product_matrix[2 * j + 1], &inc);
			delta[2 * j] += beta;
		}
	}
	delete[] alpha;
}

void NEWTON::info(const char *fmt,...)
{
	if(mpi_get_rank()!=0)
		return;
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*newton_print_string)(buf);
}

NEWTON::NEWTON(const function *fun_obj, double eps, double eta, int max_iter)
{
	this->fun_obj = const_cast<function *>(fun_obj);
	this->eps = eps;
	this->eta = eta;
	this->max_iter=max_iter;
	newton_print_string = default_print;
}

NEWTON::~NEWTON()
{
}

void NEWTON::newton(double *w)
{
	int n = fun_obj->get_nr_variable();
	int nr_node = mpi_get_size();
	int rank = mpi_get_rank();
	int shift = int(ceil(double(n) / double(nr_node)));
	int start = shift * rank;
	int length = min(max(n - start, 0), shift);
	if (length == 0)
		start = 0;
	int cg_iter;
	double cg_eps = 0.1;
	double snorm;
	double f, fnew, actred, gs;
	int search = 1, iter = 1, inc = 1;
	double *s = new double[n];
	double *r = new double[n];
	double *g = new double[n];
	double ws;

	memset(w, 0, sizeof(double) * size_t(n));

	int64_t time_st = wall_clock_ns();
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	double gnorm1 = dnrm2_(&n, g, &inc);
	double gnorm = gnorm1;
	double init_size = 1;
	double step_size;

	if (gnorm <= eps*gnorm1)
		search = 0;
	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed, time_st);
	info("iter %2d act %5.3e f %15.20e |g| %5.3e CG %3d step_size %5.3e time %5.3e\n", 0, 0.0, f, gnorm, 0, 0.0, tt);
	time_st = wall_clock_ns();

	iter = 1;
	while (iter <= max_iter && search)
	{
		cg_iter = cg(g, s, r, cg_eps);

		gs = ddot_(&length, g + start, &inc, s + start, &inc);
		snorm = ddot_(&length, s + start, &inc, s + start, &inc);
		ws = ddot_(&length, w + start, &inc, s + start, &inc);
		double buffer[3] = {gs, snorm, ws};
		mpi_allreduce(buffer, 3, MPI_DOUBLE, MPI_SUM);
		gs = buffer[0];
		snorm = sqrt(buffer[1]);
		ws = buffer[2];
		step_size = fun_obj->line_search(s, gs, snorm * snorm, ws, &fnew, eta, init_size);
		daxpy_(&n, &step_size, s, &inc, w, &inc);

		// Compute the actual reduction.
		actred = f - fnew;

		f = fnew;
		fun_obj->grad(w, g);

		gnorm = dnrm2_(&n, g, &inc);
		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		info("iter %2d act %5.3e f %15.20e |g| %5.3e CG %3d step_size %5.3e time %5.3e\n", iter, actred, f, gnorm, cg_iter, step_size, tt);
		time_st = wall_clock_ns();
		iter++;
		if (gnorm <= eps*gnorm1)
			break;
		if (f < -1.0e+32)
		{
			info("WARNING: f < -1.0e+32\n");
			break;
		}
		if (fabs(actred) <= 0)
		{
			info("WARNING: actred <= 0\n");
			break;
		}
		if (fabs(actred) <= 1.0e-12*fabs(f))
		{
			info("WARNING: actred too small\n");
			break;
		}
	}

	delete[] g;
	delete[] r;
	delete[] s;
}

int NEWTON::cg(double *g, double *s, double *r, double cg_eps)
{
	int inc = 1;
	int n = fun_obj->get_nr_variable();
	double one = 1;
	double minusone = -1;
	double *d = new double[n];
	double *Hd = new double[n + 1];
	double rTr, rnewTrnew, alpha, beta, cgtol;
	memset(s, 0, sizeof(double) * size_t(n));

	memcpy(r, g, sizeof(double) * size_t(n));
	dscal_(&n, &minusone, r, &inc);
	memcpy(d, r, sizeof(double) * size_t(n));
	cgtol = cg_eps * dnrm2_(&n, g, &inc);

	int cg_iter = 0;
	rTr = ddot_(&n, r, &inc, r, &inc);
	while (1)
	{
		if (dnrm2_(&n, r, &inc) <= cgtol)
			break;
		cg_iter++;
		fun_obj->Hv(d, Hd);

		alpha = rTr/ddot_(&n, d, &inc, Hd, &inc);
		daxpy_(&n, &alpha, d, &inc, s, &inc);
		alpha = -alpha;
		daxpy_(&n, &alpha, Hd, &inc, r, &inc);
		rnewTrnew = ddot_(&n, r, &inc, r, &inc);
		beta = rnewTrnew/rTr;
		dscal_(&n, &beta, d, &inc);
		daxpy_(&n, &one, r, &inc, d, &inc);
		rTr = rnewTrnew;
	}

	delete[] d;
	delete[] Hd;

	return(cg_iter);
}

void NEWTON::set_print_string(void (*print_string) (const char *buf))
{
	newton_print_string = print_string;
}

void TRON::info(const char *fmt,...)
{
	if(mpi_get_rank()!=0)
		return;
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*tron_print_string)(buf);
}

TRON::TRON(const function *fun_obj, double eps, double default_eps, double eps_cg, int max_iter)
{
	this->fun_obj=const_cast<function *>(fun_obj);
	this->eps=eps;
	this->eps_cg=eps_cg;
	this->max_iter=max_iter;
	this->default_eps = default_eps;
	tron_print_string = default_print;
}

TRON::~TRON()
{
}

void TRON::tron(double *w)
{
	// Parameters for updating the iterates.
	double eta0 = 1e-4, eta1 = 0.25, eta2 = 0.75;

	// Parameters for updating the trust region size delta.
	double sigma1 = 0.25, sigma2 = 0.5, sigma3 = 4;

	int n = fun_obj->get_nr_variable();
	int printed = 0;
	int nr_node = mpi_get_size();
	int rank = mpi_get_rank();
	int shift = int(ceil(double(n) / double(nr_node)));
	int start = shift * rank;
	int length = min(max(n - start, 0), shift);
	if (length == 0)
		start = 0;
	int cg_iter;
	double delta, snorm, one=1.0;
	double alpha, f, fnew, prered, actred, gs;
	int search = 1, iter = 1, inc = 1;
	double *s = new double[n];
	double *r = new double[n];
	double *g = new double[n];

	// calculate gradient norm at w=0 for stopping condition.
	memset(w, 0, sizeof(double) * size_t(n));

	int64_t time_st = wall_clock_ns();
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	delta = dnrm2_(&n, g, &inc);
	double gnorm = delta;
	double gnorm0 = gnorm;

	if (gnorm <= eps*gnorm0)
		search = 0;

	iter = 0;
	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed, time_st);
	info("iter %2d act %5.3e pre %5.3e delta %5.3e f %15.20e |g| %5.3e CG %3d time %5.3e\n", iter, 0.0, 0.0, delta, f, gnorm, 0, tt);
	time_st = wall_clock_ns();

	double *w_new = new double[n];
	while (iter <= max_iter && search)
	{
		cg_iter = trcg(delta, g, s, r);

		memcpy(w_new, w, sizeof(double) * size_t(n));
		daxpy_(&n, &one, s, &inc, w_new, &inc);

		double buffer[2];
		gs = ddot_(&length, g + start, &inc, s + start, &inc);
		buffer[0] = gs;
		buffer[1] = ddot_(&length, s + start, &inc, r + start, &inc);
		mpi_allreduce(buffer, 2, MPI_DOUBLE, MPI_SUM);
		gs = buffer[0];
		prered = -0.5*(gs-buffer[1]);
		fnew = fun_obj->fun(w_new);

		// Compute the actual reduction.
		actred = f - fnew;

		// On the first iteration, adjust the initial step bound.
		snorm = dnrm2_(&n, s, &inc);
		if (iter == 1)
			delta = min(delta, snorm);

		// Compute prediction alpha*snorm of the step.
		if (fnew - f - gs <= 0)
			alpha = sigma3;
		else
			alpha = max(sigma1, -0.5*(gs/(fnew - f - gs)));

		// Update the trust region bound according to the ratio of actual to predicted reduction.
		if (actred < eta0*prered)
			delta = min(max(alpha, sigma1)*snorm, sigma2*delta);
		else if (actred < eta1*prered)
			delta = max(sigma1*delta, min(alpha*snorm, sigma2*delta));
		else if (actred < eta2*prered)
			delta = max(sigma1*delta, min(alpha*snorm, sigma3*delta));
		else
			delta = max(delta, min(alpha*snorm, sigma3*delta));

		if (actred > eta0*prered)
		{
			iter++;
			memcpy(w, w_new, sizeof(double) * size_t(n));
			f = fnew;
			fun_obj->grad(w, g);

			gnorm = dnrm2_(&n, g, &inc);
			iter_ed = wall_clock_ns();
			tt += wall_time_diff(iter_ed, time_st);
			info("iter %2d act %5.3e pre %5.3e delta %5.3e f %15.20e |g| %5.3e CG %3d time %5.3e\n", iter, actred, prered, delta, f, gnorm, cg_iter, tt);
			time_st = wall_clock_ns();
			if (gnorm <= default_eps * gnorm0 && printed == 0)
			{
				printed = 1;
				info("Default stopping value = %15.20e\n",f);
			}
			if (gnorm <= default_eps * 0.1 * gnorm0 && printed < 2)
			{
				printed++;
				info("10 times tighter stopping value = %15.20e\n",f);
			}
			if (gnorm <= default_eps * 0.01 * gnorm0 && printed < 3)
			{
				printed++;
				info("100 times tighter stopping value = %15.20e\n",f);
			}


			if (gnorm <= eps*gnorm0)
				break;
		}
		if (f < -1.0e+32)
		{
			info("WARNING: f < -1.0e+32\n");
			break;
		}
		if (fabs(actred) <= 0 && prered <= 0)
		{
			info("WARNING: actred and prered <= 0\n");
			break;
		}
		if (fabs(actred) <= 1.0e-12*fabs(f) &&
				fabs(prered) <= 1.0e-12*fabs(f))
		{
			info("WARNING: actred and prered too small\n");
			break;
		}
	}
	if (gnorm <= default_eps * gnorm0 && printed == 0)
	{
		printed = 1;
		info("Default stopping value = %15.20e\n",f);
	}

	delete[] g;
	delete[] r;
	delete[] w_new;
	delete[] s;
}

int TRON::trcg(double delta, double *g, double *s, double *r)
{
	int inc = 1;
	int n = fun_obj->get_nr_variable();
	double one = 1;
	double minusone = -1;
	double *d = new double[n];
	double *Hd = new double[n];
	double rTr, rnewTrnew, alpha, beta, cgtol;
	memcpy(r, g, sizeof(double) * size_t(n));
	dscal_(&n, &minusone, r, &inc);
	memcpy(d, r, sizeof(double) * size_t(n));
	memset(s, 0, sizeof(double) * size_t(n));
	cgtol = eps_cg*dnrm2_(&n, g, &inc);

	int cg_iter = 0;
	rTr = ddot_(&n, r, &inc, r, &inc);
	while (1)
	{
		if (dnrm2_(&n, r, &inc) <= cgtol)
			break;
		cg_iter++;
		fun_obj->Hv(d, Hd);

		alpha = rTr/ddot_(&n, d, &inc, Hd, &inc);
		daxpy_(&n, &alpha, d, &inc, s, &inc);
		if (dnrm2_(&n, s, &inc) > delta)
		{
			info("cg reaches trust region boundary\n");
			alpha = -alpha;
			daxpy_(&n, &alpha, d, &inc, s, &inc);

			double std = ddot_(&n, s, &inc, d, &inc);
			double sts = ddot_(&n, s, &inc, s, &inc);
			double dtd = ddot_(&n, d, &inc, d, &inc);
			double dsq = delta*delta;
			double rad = sqrt(std*std + dtd*(dsq-sts));
			if (std >= 0)
				alpha = (dsq - sts)/(std + rad);
			else
				alpha = (rad - std)/dtd;
			daxpy_(&n, &alpha, d, &inc, s, &inc);
			alpha = -alpha;
			daxpy_(&n, &alpha, Hd, &inc, r, &inc);
			break;
		}
		alpha = -alpha;
		daxpy_(&n, &alpha, Hd, &inc, r, &inc);
		rnewTrnew = ddot_(&n, r, &inc, r, &inc);
		beta = rnewTrnew/rTr;
		dscal_(&n, &beta, d, &inc);
		daxpy_(&n, &one, r, &inc, d, &inc);
		rTr = rnewTrnew;
	}

	delete[] d;
	delete[] Hd;

	return(cg_iter);
}

double TRON::norm_inf(int n, double *x)
{
	double dmax = fabs(x[0]);
	for (int i=1; i<n; i++)
		if (fabs(x[i]) >= dmax)
			dmax = fabs(x[i]);
	return(dmax);
}

void TRON::set_print_string(void (*print_string) (const char *buf))
{
	tron_print_string = print_string;
}


MLBFGS::MLBFGS(const function *fun_obj, int l, double eps, int m, double eta, int max_iter):
	LBFGS(fun_obj, eps, m, eta, max_iter)
{
	this->l = l;
}

MLBFGS::~MLBFGS()
{
}

void MLBFGS::mlbfgs(double *w)
{
	int n = fun_obj->get_nr_variable();
	int nr_node = mpi_get_size();
	int rank = mpi_get_rank();
	int shift = int(ceil(double(n) / double(nr_node)));
	int start = shift * rank;
	int length = min(max(n - start, 0), shift);
	if (length == 0)
		start = 0;
	int i, k;
	double f;
	double minusone = -1;
	double one = 1;
	int search = 1, iter = 0, inc = 1;
	int DynamicM = 0;
	double **inner_product_matrix = new double*[2*M+1];
	double *g = new double[n];
	double *gx = new double[l];
	double *old_g;
	double *old_gx = NULL;
	double *step;
	double *step_x;
	double* delta = new double[M * 2 + 1];
	double **sx = new double*[M];
	double **yx = new double*[M];
	
	double **s = new double*[M];
	double **y = new double*[M];
	
	for (int i=0; i < 2*M + 1; i++)
		inner_product_matrix[i] = new double[2*M+1];
	for (int i=0; i < M;i++)
		y[i] = new double[length];
	double* linear = new double[2 * M + 1];
	double *C_times_D = new double[l];
	double gnorm, gnorm1;
	double step_size = 0;
	double origDirDeriv = 0;
	double err = 0;
	double dTd;

	memset(w, 0, sizeof(double) * size_t(n));
	int64_t time_st = wall_clock_ns();
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	fun_obj->Xv(g, gx);
	gnorm1 = dnrm2_(&n, g, &inc);
	gnorm = gnorm1;

	if (gnorm <= eps*gnorm1)
		search = 0;

	inner_product_matrix[0][0] = gnorm * gnorm;

	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed,time_st);
	int tmp_length = 0;
	int previous_M = 0;
	time_st = wall_clock_ns();
	k = 0;
	double *system = new double[(2 * M + 1) * (2 * M + 1)];

	while (iter < max_iter && search)
	{
		fun_obj->Hessian_coeff(C_times_D);
		gnorm = sqrt(inner_product_matrix[2 * DynamicM][2 * DynamicM]);
		info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
		if (gnorm <= eps*gnorm1)
			break;

		memset(system, 0, sizeof(double) * size_t((2 * M + 1) * (2 * M + 1)));
		compute_system(iter, k, n, DynamicM, inner_product_matrix, sx, yx, gx, C_times_D, system);
		
		if (iter == 0)
			k--;
		k = (k+1)%M;

		memset(delta, 0, sizeof(double) * size_t(M * 2 + 1));
		tmp_length = 2 * DynamicM + 1;
		memcpy(linear, inner_product_matrix[2 * DynamicM], sizeof(double) * size_t(tmp_length));
		if (iter <= M)
			for (i=0;i<DynamicM;i++)
				linear[2 * i] = 0;

		dscal_(&tmp_length, &minusone, linear, &inc);
		int system_size = 2 * DynamicM + 1;
		err = solve_system(system, linear, 2 * DynamicM + 1, system_size, delta);

		step = new double[length];
		step_x = new double[l];
		memcpy(step, g + start, sizeof(double) * size_t(length));
		memcpy(step_x, gx, sizeof(double) * size_t(l));
		dscal_(&length, delta + (2 * DynamicM), step, &inc);
		dscal_(&l, delta + (2 * DynamicM), step_x, &inc);

		for (i=0;i<DynamicM;i++)
		{
			daxpy_(&length, delta + (2 * i + 1), y[i], &inc, step, &inc);
			daxpy_(&l, delta + (2 * i + 1), yx[i], &inc, step_x, &inc);
		}

		if (iter > M)
			for (i=0;i<DynamicM;i++)
			{
				daxpy_(&length, delta + (2 * i), s[i], &inc, step, &inc);

				daxpy_(&l, delta + (2 * i), sx[i], &inc, step_x, &inc);
			}

		origDirDeriv = -ddot_(&tmp_length, delta, &inc, linear, &inc);
		dTd = 0;
		for (i=0;i<2 * DynamicM + 1; i++)
			dTd += delta[i] * ddot_(&tmp_length, delta, &inc, inner_product_matrix[i], &inc);
		double ws = ddot_(&length, w + start, &inc, step, &inc);
		mpi_allreduce(&ws, 1, MPI_DOUBLE, MPI_SUM);
		step_size = fun_obj->line_search(step, origDirDeriv, dTd, ws, &f, eta, 1.0, step_x);
		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}
		dscal_(&length, &step_size, step, &inc);
		dscal_(&l, &step_size, step_x, &inc);
		daxpy_(&length, &one, step, &inc, w + start, &inc);
		int64_t iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		time_st = wall_clock_ns();
		old_g = g;
		old_gx = gx;
		iter++;
		previous_M = DynamicM;
		DynamicM = min(iter, M);

		if (iter > M)
		{
			delete[] s[k];
			delete[] sx[k];
			delete[] yx[k];
		}

		s[k] = step;
		sx[k] = step_x;
		memcpy(y[k], old_g + start, sizeof(double) * size_t(length));
		yx[k] = old_gx;
		delete[] old_g;
		g = new double[n];
		fun_obj->grad(w, g);
		gx = new double[l];
		fun_obj->Xv(g, gx);

		daxpy_(&length, &minusone, g + start, &inc, y[k], &inc);
		daxpy_(&l, &minusone, gx, &inc, yx[k], &inc);
		update_inner_products(inner_product_matrix, k, n, DynamicM, previous_M, s, y, g, delta, step_size);
	}

	delete[] delta;
	delete[] system;
	for (i=0;i<DynamicM;i++)
	{
		delete[] sx[i];
		delete[] s[i];
		delete[] yx[i];
		delete[] y[i];
	}
	delete[] sx;
	delete[] yx;
	delete[] y;
	delete[] s;
	delete[] g;
	delete[] gx;
	delete[] C_times_D;
	delete[] linear;

	int *recv_count = new int[nr_node];
	int *displace = new int[nr_node];
	int counter = 0;
	for (i=0;i<nr_node;i++)
	{
		recv_count[i] = shift;
		displace[i] = counter;
		counter += shift;
		if (counter >= n)
		{
			counter = 0;
			shift = 0;
		}
		else if (counter + shift > n)
			shift = n - counter;
	}
	double *recv_buffer = new double[n];
	MPI_Allgatherv(w + start, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
	memcpy(w, recv_buffer, sizeof(double) * size_t(n));

	delete[] recv_count;
	delete[] recv_buffer;
	delete[] displace;

}

void MLBFGS::compute_system(int iter, int k, int n, int DynamicM, double** inner_product_matrix, double** sx, double** yx, double* gx, double* C_times_D, double* system)
{
	int i,j;
	int length = (2 * DynamicM + 1) * (2 * DynamicM + 2) / 2;
	int counter = 0;

	if (iter <= DynamicM)
		length = (DynamicM + 2) * (DynamicM + 1) / 2;

	double* buffer = new double[length];
	double entry;

	if (iter <= DynamicM)
	{
		for (i=0;i<DynamicM;i++)
		{
			for (j=0;j<=i;j++)
			{
				buffer[counter] = ddot3(yx[i], C_times_D, yx[j], l);
				counter++;
			}
		}
		for (i=0;i<DynamicM;i++)
		{
			buffer[counter] = ddot3(gx, C_times_D, yx[i], l);
			counter++;
		}
	}
	else
	{
		for (i=0;i<DynamicM;i++)
		{
			for (j=0;j<i;j++)
			{
				buffer[counter] = ddot3(sx[i], C_times_D, sx[j], l);
				counter++;
				buffer[counter] = ddot3(sx[i], C_times_D, yx[j], l);
				counter++;
			}
			buffer[counter] = ddot3(sx[i], C_times_D, sx[i], l);
			counter++;

			for (j=0;j<=i;j++)
			{
				buffer[counter] = ddot3(yx[i], C_times_D, sx[j], l);
				counter++;
				buffer[counter] = ddot3(yx[i], C_times_D, yx[j], l);
				counter++;
			}
		}
		for (i=0;i<DynamicM;i++)
		{
			buffer[counter] = ddot3(gx, C_times_D, sx[i], l);
			counter++;
			buffer[counter] = ddot3(gx, C_times_D, yx[i], l);
			counter++;
		}
	}
	buffer[counter] = ddot3(gx, C_times_D, gx, l);
	counter++;

	mpi_allreduce(buffer, counter, MPI_DOUBLE, MPI_SUM);
	counter = 0;
	if (iter <= DynamicM)
	{
		for (i=0;i<DynamicM;i++)
			for (j=0;j<=i;j++)
			{
				entry = buffer[counter] + inner_product_matrix[2 * i + 1][2 * j + 1];
				counter++;
				system[(2 * i + 1) * (2 * DynamicM + 1) + 2 * j + 1] = entry;
				system[(2 * j + 1) * (2 * DynamicM + 1) + 2 * i + 1] = entry;
			}
		for (i=0;i<DynamicM;i++)
		{
			memset(system + (2 * i) * (2 * DynamicM + 1), 0, sizeof(double) * size_t(2 * DynamicM + 1));
			for (j=0;j<DynamicM;j++)
				system[(2 * j + 1) * (2 * DynamicM + 1) + 2 * i] = 0;
			system[(2 * DynamicM) * (2 * DynamicM + 1) + 2 * i] = 0;
		}
		for (i=0;i<DynamicM;i++)
		{
			entry = buffer[counter] + inner_product_matrix[2 * DynamicM][2 * i + 1];
			counter++;
			system[(2 * DynamicM) * (2 * DynamicM + 1) + 2 * i + 1] = entry;
			system[(2 * i + 1) * (2 * DynamicM + 1) + 2 * DynamicM] = entry;
		}
	}
	else
	{
		for (i=0;i<DynamicM;i++)
		{
			for (j=0;j<i;j++)
			{
				entry = buffer[counter] + inner_product_matrix[2 * i][2 * j];
				counter++;
				system[2 * i * (2 * DynamicM + 1) + 2 * j] = entry;
				system[2 * j * (2 * DynamicM + 1) + 2 * i] = entry;
				entry = buffer[counter] + inner_product_matrix[2 * i][2 * j + 1];
				counter++;
				system[2 * i * (2 * DynamicM + 1) + 2 * j + 1] = entry;
				system[(2 * j + 1) * (2 * DynamicM + 1) + 2 * i] = entry;
			}
			system[2 * i * (2 * DynamicM + 1) + 2 * i] = buffer[counter] + inner_product_matrix[2 *i][2 * i];
			counter++;

			for (j=0;j<=i;j++)
			{
				entry = buffer[counter] + inner_product_matrix[2 * i + 1][2 * j];
				counter++;
				system[(2 * i + 1) * (2 * DynamicM + 1) + 2 * j] = entry;
				system[(2 * j) * (2 * DynamicM + 1) + 2 * i + 1] = entry;
				entry = buffer[counter] + inner_product_matrix[2 * i + 1][2 * j + 1];
				counter++;
				system[(2 * i + 1) * (2 * DynamicM + 1) + 2 * j + 1] = entry;
				system[(2 * j + 1) * (2 * DynamicM + 1) + 2 * i + 1] = entry;
			}
		}

		for (i=0;i<DynamicM;i++)
		{
			entry = buffer[counter] + inner_product_matrix[2 * DynamicM][2 * i];
			counter++;
			system[(2 * DynamicM) * (2 * DynamicM + 1) + 2 * i] = entry;
			system[(2 * i) * (2 * DynamicM + 1) + 2 * DynamicM] = entry;
			entry = buffer[counter] + inner_product_matrix[2 * DynamicM][2 * i + 1];
			counter++;
			system[(2 * DynamicM) * (2 * DynamicM + 1) + 2 * i + 1] = entry;
			system[(2 * i + 1) * (2 * DynamicM + 1) + 2 * DynamicM] = entry;
		}
	}
	system[(2 * DynamicM + 1) * (2 * DynamicM + 1) - 1] = buffer[counter] + inner_product_matrix[2 * DynamicM][2 * DynamicM];
	
	delete[] buffer;
}

void MLBFGS::update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, int previous_M, double** s, double** y, double* g, double* delta, double step_size)
{
	int i;
	int prev_length = 2 * previous_M + 1;
	int nr_node = mpi_get_size();
	int rank = mpi_get_rank();
	int shift = int(ceil(double(n) / double(nr_node)));
	int start = shift * rank;
	int length = min(max(n - start, 0), shift);
	int inc = 1;
	if (length == 0)
		start = 0;
	double *tmp = new double[prev_length + 1];
	double *tmp2 = new double[2 * DynamicM];

	double* buffer = new double[DynamicM * 2 + 2];
	memset(buffer, 0, sizeof(double) * size_t(DynamicM * 2 + 2));
	for (i=0;i<prev_length;i++)
		tmp[i] = step_size * ddot_(&prev_length, delta, &inc, inner_product_matrix[i], &inc);
	tmp[prev_length] = step_size * ddot_(&prev_length, delta, &inc, tmp, &inc);

	for (i=0;i<DynamicM;i++)
	{
		buffer[2 * i] = ddot_(&length, g + start, &inc, s[i], &inc);
		buffer[2 * i + 1] = ddot_(&length, g + start, &inc, y[i], &inc);
	}
	buffer[2 * DynamicM] = ddot_(&length, g + start, &inc, g + start, &inc);
	buffer[2 * DynamicM + 1] = ddot_(&length, y[k], &inc, y[k], &inc);
	mpi_allreduce(buffer, 2 * DynamicM + 2, MPI_DOUBLE, MPI_SUM);
	for (i=0;i<prev_length - 1; i++)
		tmp2[i] = inner_product_matrix[prev_length-1][i] - buffer[i] ;

	tmp2[2 * k] = tmp[prev_length - 1] - buffer[2 * k];
//	s[k] * y[k] = s[k] * g - s[k] * old_g

	tmp2[2 * k + 1] = buffer[2 * DynamicM + 1];
//	y[k] * y[k]

	for (i=0;i< prev_length - 1;i++)
	{
		inner_product_matrix[2 * k][i] = tmp[i];
		inner_product_matrix[i][2 * k] = tmp[i];
	}

	for (i=0;i< 2 * DynamicM;i++)
	{
		inner_product_matrix[2 * k + 1][i] = tmp2[i];
		inner_product_matrix[i][2 * k + 1] = tmp2[i];
	}

	inner_product_matrix[2 * k][2 * k] = tmp[prev_length];

	for (i=0; i< 2 * DynamicM + 1; i++)
	{
		inner_product_matrix[2 * DynamicM][i] = buffer[i];
		inner_product_matrix[i][2 * DynamicM] = buffer[i];
	}

	delete[] buffer;
	delete[] tmp;
	delete[] tmp2;
}

MG::MG(const function *fun_obj, int l, double eps, int m, double eta, int max_iter)
{
	this->fun_obj=const_cast<function *>(fun_obj);
	this->eps=eps;
	this->max_iter=max_iter;
	this->eta = eta;
	mg_print_string = default_print;
	this->M = m;
	this->l = l;
}

MG::~MG()
{
}

void MG::mg(double *w)
{
	int n = fun_obj->get_nr_variable();
	int nr_node = mpi_get_size();
	int rank = mpi_get_rank();
	int shift = int(ceil(double(n) / double(nr_node)));
	int start = shift * rank;
	int length = min(max(n - start, 0), shift);
	if (length == 0)
		start = 0;
	int i, k;
	double f;
	double minusone = -1;
	int search = 1, iter = 0, inc = 1;
	int DynamicM = 0;
	double **inner_product_matrix = new double*[M];
	double *g = new double[n];
	double* step = new double[length];
	double* step_x = new double[l];
	double* delta = new double[M];
	double **px = new double*[M];
	
	double **p = new double*[M];
	
	for (int i=0; i < M; i++)
		inner_product_matrix[i] = new double[M];
	double* linear = new double[M];
	double *C_times_D = new double[l];
	double gnorm, gnorm1;
	double step_size = 0;
	double origDirDeriv = 0;
	double err = 0;
	double dTd;

	memset(w, 0, sizeof(double) * size_t(n));
	int64_t time_st = wall_clock_ns();
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	gnorm1 = dnrm2_(&n, g, &inc);
	gnorm = gnorm1;

	if (gnorm <= eps*gnorm1)
		search = 0;
	inner_product_matrix[0][0] = gnorm * gnorm;

	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed,time_st);
	int tmp_length = 0;
	time_st = wall_clock_ns();
	info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter, 0, gnorm, f, 0.0, 0.0, tt);
	k = 0;

	iter = 0;
	while (iter < max_iter && search)
	{
		DynamicM = min(iter + 1, M);
		if (iter > M)
		{
			delete[] p[k];
			delete[] px[k];
		}
		px[k] = new double[l];
		p[k] = new double[length];
		memcpy(p[k], g + start, sizeof(double) * size_t(length));
		if (iter != 0)
			update_inner_products(inner_product_matrix, k, n, DynamicM, p, g);
		fun_obj->Xv(g, px[k]);
		gnorm = sqrt(inner_product_matrix[k][k]);
		if (gnorm <= eps*gnorm1)
			break;
		fun_obj->Hessian_coeff(C_times_D);
		
		double *system = new double[DynamicM * DynamicM];
		compute_system(iter, k, n, DynamicM, inner_product_matrix, px, C_times_D, system);
		
		memset(delta, 0, sizeof(double) * size_t(M));
		tmp_length = DynamicM;
		memcpy(linear, inner_product_matrix[k], sizeof(double) * size_t(tmp_length));
		dscal_(&tmp_length, &minusone, linear, &inc);
		int system_size = DynamicM;
		err = solve_system(system, linear, DynamicM, system_size, delta);
		memset(step, 0, sizeof(double) * size_t(length));
		memset(step_x, 0, sizeof(double) * size_t(l));

		for (i=0;i<DynamicM;i++)
		{
			daxpy_(&length, delta + i, p[i], &inc, step, &inc);
			daxpy_(&l, delta + i, px[i], &inc, step_x, &inc);
		}

		origDirDeriv = -ddot_(&tmp_length, delta, &inc, linear, &inc);
		dTd = 0;
		for (i=0;i<DynamicM; i++)
			dTd += delta[i] * ddot_(&tmp_length, delta, &inc, inner_product_matrix[i], &inc);
		delete[] system;
		double ws = ddot_(&length, w + start, &inc, step, &inc);
		mpi_allreduce(&ws, 1, MPI_DOUBLE, MPI_SUM);
		step_size = fun_obj->line_search(step, origDirDeriv, dTd, ws, &f, eta, 1.0, step_x);
		info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter + 1,DynamicM, gnorm, f, step_size, err, tt);
		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}
		daxpy_(&length, &step_size, step, &inc, w + start, &inc);
		int64_t iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		time_st = wall_clock_ns();
		k = (k+1)%M;
		iter++;
		fun_obj->grad(w, g);
	}
	delete[] step;
	delete[] step_x;
	delete[] delta;
	int *recv_count = new int[nr_node];
	int *displace = new int[nr_node];
	int counter = 0;
	for (i=0;i<nr_node;i++)
	{
		recv_count[i] = shift;
		displace[i] = counter;
		counter += shift;
		if (counter >= n)
		{
			counter = 0;
			shift = 0;
		}
		else if (counter + shift > n)
			shift = n - counter;
	}
	double *recv_buffer = new double[n];
	MPI_Allgatherv(w + start, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
	memcpy(w, recv_buffer, sizeof(double) * size_t(n));

	delete[] recv_count;
	delete[] recv_buffer;
	delete[] displace;

	for (i=0;i<DynamicM;i++)
	{
		delete[] px[i];
		delete[] p[i];
	}
	delete[] px;
	delete[] p;
	delete[] g;
	delete[] C_times_D;
	delete[] linear;
}

MP::MP(const function *fun_obj, int l, double eps, int m, double eta, int max_iter)
	: MG(fun_obj, l, eps, m, eta, max_iter)
{
}

void MP::mp(double *w)
{
	int n = fun_obj->get_nr_variable();
	int nr_node = mpi_get_size();
	int rank = mpi_get_rank();
	int shift = int(ceil(double(n) / double(nr_node)));
	int start = shift * rank;
	int length = min(max(n - start, 0), shift);
	if (length == 0)
		start = 0;
	int i, k;
	double f;
	double minusone = -1;
	int search = 1, iter = 0, inc = 1;
	int DynamicM = 0;
	int previous_M = 0;
	double **inner_product_matrix = new double*[M];
	double *g = new double[n];
	double *step;
	double *step_x;
	double* delta = new double[M];
	double **px = new double*[M];
	
	double **p = new double*[M];
	
	for (int i=0; i < M; i++)
		inner_product_matrix[i] = new double[M];
	double* linear = new double[M];
	double *C_times_D = new double[l];
	double gnorm, gnorm1;
	double step_size = 0;
	double origDirDeriv = 0;
	double err = 0;
	double dTd;

	memset(w, 0, sizeof(double) * size_t(n));
	int64_t time_st = wall_clock_ns();
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	gnorm1 = dnrm2_(&n, g, &inc);
	gnorm = gnorm1;

	if (gnorm <= eps*gnorm1)
		search = 0;
	inner_product_matrix[0][0] = gnorm * gnorm;

	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed,time_st);
	int tmp_length = 0;
	time_st = wall_clock_ns();
	info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter, 0, gnorm, f, 0.0, 0.0, tt);
	k = 0;

	iter = 0;
	while (iter < max_iter && search)
	{
		DynamicM = min(iter + 1, M);
		if (iter > M)
		{
			delete[] p[k];
			delete[] px[k];
		}
		px[k] = new double[l];
		p[k] = g + start;
		if (iter != 0)
			update_inner_products(inner_product_matrix, k, n, DynamicM, previous_M, p, g, delta, step_size);
		fun_obj->Xv(g, px[k]);
		gnorm = sqrt(inner_product_matrix[k][k]);
		if (gnorm <= eps*gnorm1)
			break;
		fun_obj->Hessian_coeff(C_times_D);
		
		double *system = new double[DynamicM * DynamicM];
		compute_system(iter, k, n, DynamicM, inner_product_matrix, px, C_times_D, system);
		
		memset(delta, 0, sizeof(double) * size_t(M));
		tmp_length = DynamicM;
		memcpy(linear, inner_product_matrix[k], sizeof(double) * size_t(tmp_length));
		dscal_(&tmp_length, &minusone, linear, &inc);
		int system_size = DynamicM;
		err = solve_system(system, linear, DynamicM, system_size, delta);
		step = new double[length];
		step_x = new double[l];
		memset(step, 0, sizeof(double) * size_t(length));
		memset(step_x, 0, sizeof(double) * size_t(l));

		for (i=0;i<DynamicM;i++)
		{
			daxpy_(&length, delta + i, p[i], &inc, step, &inc);
			daxpy_(&l, delta + i, px[i], &inc, step_x, &inc);
		}

		origDirDeriv = -ddot_(&tmp_length, delta, &inc, linear, &inc);
		dTd = 0;
		for (i=0;i<DynamicM; i++)
			dTd += delta[i] * ddot_(&tmp_length, delta, &inc, inner_product_matrix[i], &inc);
		delete[] system;
		double ws = ddot_(&length, w + start, &inc, step, &inc);
		mpi_allreduce(&ws, 1, MPI_DOUBLE, MPI_SUM);
		step_size = fun_obj->line_search(step, origDirDeriv, dTd, ws, &f, eta, 1.0, step_x);
		info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter + 1,DynamicM, gnorm, f, step_size, err, tt);
		delete[] px[k];
		p[k] = step;
		px[k] = step_x;
		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}
		daxpy_(&length, &step_size, step, &inc, w + start, &inc);
		int64_t iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		time_st = wall_clock_ns();
		iter++;
		previous_M = DynamicM;
		k = (k+1)%M;
		fun_obj->grad(w, g);
	}
	delete[] delta;
	int *recv_count = new int[nr_node];
	int *displace = new int[nr_node];
	int counter = 0;
	for (i=0;i<nr_node;i++)
	{
		recv_count[i] = shift;
		displace[i] = counter;
		counter += shift;
		if (counter >= n)
		{
			counter = 0;
			shift = 0;
		}
		else if (counter + shift > n)
			shift = n - counter;
	}
	double *recv_buffer = new double[n];
	MPI_Allgatherv(w + start, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
	memcpy(w, recv_buffer, sizeof(double) * size_t(n));

	delete[] recv_count;
	delete[] recv_buffer;
	delete[] displace;

	if (iter < M)
		DynamicM = k;
	for (i=0;i<DynamicM;i++)
	{
		delete[] px[i];
		delete[] p[i];
	}
	delete[] px;
	delete[] p;
	delete[] g;
	delete[] C_times_D;
	delete[] linear;
	for (i=0;i<M;i++)
		delete[] inner_product_matrix[i];
	delete[] inner_product_matrix;
}

void MG::update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, double** p, double *g)
{
	int i;
	int nr_node = mpi_get_size();
	int rank = mpi_get_rank();
	int shift = int(ceil(double(n) / double(nr_node)));
	int start = shift * rank;
	int length = min(max(n - start, 0), shift);
	int inc = 1;
	if (length == 0)
		start = 0;

	double* buffer = new double[DynamicM];
	memset(buffer, 0, sizeof(double) * size_t(DynamicM));


	for (i=0;i<DynamicM;i++)
		buffer[i] = ddot_(&length, g + start, &inc, p[i], &inc);
	mpi_allreduce(buffer, DynamicM, MPI_DOUBLE, MPI_SUM);

	for (i=0; i < DynamicM; i++)
	{
		inner_product_matrix[k][i] = buffer[i];
		inner_product_matrix[i][k] = buffer[i];
	}

	delete[] buffer;
}

void MP::update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, int previous_M, double** p, double *g, double *delta, double step_size)
{
	int i;
	int inc = 1;
	int prev_length = previous_M;
	double *tmp = new double[prev_length];

	int prev_k = (k - 1 + DynamicM) % DynamicM;
	for (i=0;i<prev_length;i++)
		tmp[i] = ddot_(&prev_length, delta, &inc, inner_product_matrix[i], &inc);
	double temp = ddot_(&prev_length, delta, &inc, tmp, &inc);
	tmp[prev_k] = temp;

	for (i=0; i < k; i++)
	{
		inner_product_matrix[prev_k][i] = tmp[i];
		inner_product_matrix[i][prev_k] = tmp[i];
	}
	for (i=k + 1;i<DynamicM;i++)
	{
		inner_product_matrix[prev_k][i] = tmp[i];
		inner_product_matrix[i][prev_k] = tmp[i];
	}
	MG::update_inner_products(inner_product_matrix, k, n, DynamicM, p, g);
}

void MG::compute_system(int iter, int k, int n, int DynamicM, double** inner_product_matrix, double** px, double* C_times_D, double* system)
{
	int i,j;
	int length = DynamicM * (DynamicM + 1) / 2;
	double* buffer = new double[length];
	
	int counter = 0;
	double entry;

	for (i=0;i<DynamicM;i++)
		for (j=0;j<=i;j++)
		{
			buffer[counter] = ddot3(px[i], C_times_D, px[j], l);
			counter++;
		}
	mpi_allreduce(buffer, counter, MPI_DOUBLE, MPI_SUM);
	counter = 0;
	for (i=0;i<DynamicM;i++)
		for (j=0;j<=i;j++)
		{
			entry = buffer[counter] + inner_product_matrix[i][j];
			counter++;
			system[i * DynamicM + j] = entry;
			system[j * DynamicM + i] = entry;
		}
	delete[] buffer;
}

void MG::info(const char *fmt,...)
{
	if(mpi_get_rank()!=0)
		return;
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*mg_print_string)(buf);
}

void MG::set_print_string(void (*print_string) (const char *buf))
{
	mg_print_string = print_string;
}
